
<div style="height: 24pt"><p/></div>

<button class="accordion"><i>About this comparison</i></button>
<div class="panel">
<p>About Z-scores in this analysis:</p>
<ul>
<li>Positive Z-scores mean better than average, negative Z-scores mean worse than average.</li>
<li>"Average" is determined over a bag of models of varying sizes, updated periodically. <a href="https://github.com/leondz/garak/blob/main/garak/resources/calibration/bag.md">Details</a></li>
<li>For any probe, roughly two-thirds of models get a Z-score between -1.0 and +1.0.</li>
<li>The middle 10% of models score -0.125 to +0.125. This is labelled "competitive".</li>
<li>A Z-score of +1.0 means the score was one standard deviation better than the mean score other models achieved for this probe &amp; metric</li>
{%if calibration_date != ""%}
<li>This run was produced using a calibration over {{model_count}} models, built at {{calibration_date}}</li>
<li>Model reports used: <i>{{model_list}}</i>.</li>
{%endif%}
</ul>
</div>


<script>
var acc = document.getElementsByClassName("accordion");
var i;

for (i = 0; i < acc.length; i++) {
  acc[i].addEventListener("click", function() {
    /* Toggle between adding and removing the "active" class,
    to highlight the button that controls the panel */
    this.classList.toggle("active");

    /* Toggle between hiding and showing the active panel */
    var panel = this.nextElementSibling;
    if (panel.style.display === "block") {
      panel.style.display = "none";
    } else {
      panel.style.display = "block";
    }
  });
}</script>

<p style="height:400px; padding-top: 100px">
generated with <a href="https://garak.ai">garak</a>
</p>

</body>

</html>