# SPDX-FileCopyrightText: Portions Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

PROBE_RGB = "e5a70e"
DETECTOR_RGB = "2c79da"
GENERATOR_RGB = "c061cb"
